#include "cpu.h"
#include "stages.h"
#include <cstdio>
#include <cassert>

const uint32_t CACHE_MISS_NUMBER_OF_STALLS = 4;			// 4 stall cycles for cache miss (5 cycles)


void CacheCtrl::cycle()
{
	if (!requestQueue.empty())
	{
		assert(requestQueue.size()==1);			//in this simple pipeline simulator, we can only have one request from MEM stage.
		
		//<CAR-PA2-REF> reference code for handling access requests
		//
		//CachePacket* pkt = requestQueue.front();
		//
		//CacheBlk *blk=cache->access(pkt->addr);
		//
		//if (blk && blk->isValid()) {
		//	responseToCPU(pkt);
		//	hits++;
		//	requestQueue.pop_front();
		//	
		//} else {
		//	//If there is a pendingMissReq, we should retry it in the next cycle.
		//	//This will not happen before you implement the prefetcher.			
		//	if (pendingMissReq == NULL) {		//only support one outstanding miss request
		//		//handle cache miss				
		//		pendingMissReq=pkt;
		//		pendingMissReq->readyCycleCount += CACHE_MISS_NUMBER_OF_STALLS;
		//		
		//		requestQueue.pop_front();
		//		misses++;
		//	}
		//}
	}

	if (pendingMissReq && (core->cycles >= pendingMissReq->readyCycleCount) ) {	
		
		handleCacheFill(pendingMissReq);
		responseToCPU(pendingMissReq);
		
		pendingMissReq=NULL;
	}
}

void CacheCtrl::access(CachePacket *pkt)
{	
	//if (cache) {
	//	requestQueue.push_back(pkt);
	//} else {					//no cache, always miss.
		pendingMissReq=pkt;
		pendingMissReq->readyCycleCount += CACHE_MISS_NUMBER_OF_STALLS;
		misses++;
	//}
}

void CacheCtrl::handleCacheFill(CachePacket *pkt)
{	
	//<CAR-PA2-REF> reference code for handling cache fill
	//
	//CacheBlk *victimBlk=cache->findVictim(pkt->addr);
	//
	//if (victimBlk->isValid() && victimBlk->isDirty()) {
	//	//Theoreically, cache will issue a writeback request to the memory, 
	//	//but you can assume ignore this writeback here.
	//	//issueWriteBack();
	//}
	//
	//victimBlk=cache->insertBlock(pkt->addr,victimBlk);
	//
	//if (pkt->reqType == CacheWriteReq)
	//	victimBlk->markDirty();	
}

void CacheCtrl::responseToCPU(CachePacket *pkt)
{
	//NOTE: You do not need to simulate the data in the cache.
	//This is a hack for you to get the data from memory.
	if (pkt->reqType == CacheWriteReq) {
		if (pkt->size == 1) {
			core->mem->set<byte>(pkt->addr, pkt->data);
		} else if (pkt->size == 4) {
			core->mem->set<uint32_t>(pkt->addr, pkt->data);
		} else {
			assert(1);
		}
	} else 	if (pkt->reqType == CacheReadReq) {
		if (pkt->size == 1) {
			pkt->data = core->mem->get<byte>(pkt->addr);
		} else if (pkt->size == 4) {
			pkt->data = core->mem->get<uint32_t>(pkt->addr);
		} else {
			assert(1);
		}
	} else {
		assert(1);
	}
	
	core->mys.response(pkt);
}

void CacheCtrl::printStat()
{
	//print the statistic here
	printf("stat.dcache_hits: %d\n", hits);
	printf("stat.dcache_misses: %d\n", misses);
	printf("stat.dcache_miss_rate: %.2f %%\n", ((double)misses/(double)(hits+misses)) * 100);	
}